/*
    Project: TCTEMP
    Version: V1.7
    File   : na_main.c

    Copyright (c) 2006-2008 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "na.h"
#include "na_image.h"
#include "na_tc.h"
#include "tc.h"

#define SYSTEM_INFORMATION_SIZE 0x10000

typedef struct
{
    TC_ENTROPY_POOL ep;
} ENTROPY_CONTEXT, *PENTROPY_CONTEXT;

typedef struct
{
    /* the image context must be the first field due to alignment issues */

    IMAGE_CONTEXT      ic;

    BYTE               rgbSystemInformation[SYSTEM_INFORMATION_SIZE];

    PCWSTR             pcwszVolumeName;

    TC_CONTEXT         tc;

    TC_HEADER          tch;
    TC_BUFFER          tcb;
    TC_PASSWORD        tcpw;

    ENTROPY_CONTEXT    ec;

} TCTEMP_CONTEXT, *PTCTEMP_CONTEXT;


/*----------------------------------------------------------------------------*/
/* Entropy                                                                    */

#define ES_SYSTEM_INFO(iSI) 0
#define ES_BASIC_FILE_INFO  0
#define ES_TSC              0
#define ES_VIA_RNG          0

#define NaFillEntropyPool(pec, iEntropySource, pData, cbDataSize) \
  TC_FillEntropyPool(&pec->ep, pData, cbDataSize);

typedef struct
{
    FILE_BASIC_INFORMATION fbi;
    WCHAR                  wszNtFileName[MAX_NT_PATH];

} FILE_INFO_LOCALS, *PFILE_INFO_LOCALS;

static BOOL NAAPI NaFillEntropyPoolWithFileInfo(PENTROPY_CONTEXT pec,
  PCWSTR pwszFileName, PVOID pBuffer, SIZE_T cbBufferSize)
{
    PFILE_INFO_LOCALS pfil;
    HANDLE            hFile;
    NTSTATUS          ntStatus;
    BOOL              fResult;

    if (cbBufferSize < sizeof(FILE_INFO_LOCALS)) {
        return FALSE;
    }

    pfil = (PFILE_INFO_LOCALS)pBuffer;
    if (!NaGetNtFileName(pfil->wszNtFileName, countof(pfil->wszNtFileName),
      pwszFileName)) {
        return FALSE;
    }

    fResult = FALSE;
    if (NaOpenFile(&hFile, pfil->wszNtFileName, FILE_READ_ATTRIBUTES,
      FILE_SHARE_READ|FILE_SHARE_WRITE, 0)) {
        if (NaQueryFileInfo(hFile, &pfil->fbi, sizeof(pfil->fbi),
          FileBasicInformation)) {
            NaFillEntropyPool(pec, ES_BASIC_FILE_INFO, &pfil->fbi,
              sizeof(pfil->fbi));
            fResult = TRUE;
        }
        NaCloseFile(hFile);
    }

    return TRUE;
}

static void NAAPI NaFillEntropyPoolWithSystemInformation(PENTROPY_CONTEXT pec,
  PVOID pBuffer, ULONG ulBufferSize)
{
    static const int rgiSI[] = {
        2, 3, 5, 6, 8, 11, 16, 21, 23, 33, 36, 43, 45, 51, 55, 0
    };

    ULONG     ulReturnLength;
    const int *pi;

    for (pi = rgiSI; *pi; pi++) {
        if (NaGetSystemInformation(*pi, pBuffer, ulBufferSize,
          &ulReturnLength)) {
            NaFillEntropyPool(pec, ES_SYSTEM_INFO(*pi), pBuffer,
              ulReturnLength);
        }
    }
}

extern unsigned __int64 rdtsc(void);

static void NAAPI NaFillEntropyPoolWithTimeStampCounter(PENTROPY_CONTEXT pec)
{
    unsigned __int64 ull;

    ull = rdtsc();
    NaFillEntropyPool(pec, ES_TSC, &ull, sizeof(ull));
}

extern int get_rng_bytes(unsigned __int8 *pbBuffer, size_t cbBufferSize);

static void NAAPI NaFillEntropyPoolWithHardwareRNG(PENTROPY_CONTEXT pec,
  PVOID pBuffer, ULONG ulBufferSize)
{
    if (get_rng_bytes((unsigned __int8 *)pBuffer, ulBufferSize)) {
        NaFillEntropyPool(pec, ES_VIA_RNG, pBuffer, ulBufferSize);
    }
}

/*----------------------------------------------------------------------------*/
/* Volume header                                                              */

static BOOL NAAPI NaReadVolumeHeader(PTCTEMP_CONTEXT pCtx)
{
    BOOL fResult;

    if (!NaGetNtFileName(pCtx->ic.wszNtImageFileName, countof(pCtx->ic.wszNtImageFileName),
      pCtx->pcwszVolumeName)) {
        return FALSE;
    }

    fResult = FALSE;
    if (NaOpenFile(&pCtx->ic.hHeader, pCtx->ic.wszNtImageFileName, GENERIC_READ,
      FILE_SHARE_READ, FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {
        LARGE_INTEGER li;
        li.QuadPart = 0;
        if (NaReadFile(pCtx->ic.hHeader, &pCtx->tch, sizeof(pCtx->tch), NULL,
          &li)) {
            fResult = TRUE;
        }
        NaCloseFile(pCtx->ic.hHeader);
    }

    return fResult;
}

static BOOL NAAPI NaWriteVolumeHeader(PTCTEMP_CONTEXT pCtx)
{
    BOOL fResult;

    fResult = FALSE;
    if (NaOpenFile(&pCtx->ic.hHeader, pCtx->ic.wszNtImageFileName,
      GENERIC_WRITE, 0, FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT)) {
        LARGE_INTEGER li;
        li.QuadPart = 0;
        if (NaWriteFile(pCtx->ic.hHeader, &pCtx->tch, sizeof(pCtx->tch), NULL,
          &li)) {
            fResult = TRUE;
        }
        NaCloseFile(pCtx->ic.hHeader);
    }

    return fResult;
}


/*----------------------------------------------------------------------------*/
/* Stack                                                                      */

void NAAPI NaWipeStack(void)
{
    BYTE rgb[2048];

    RtlSecureZeroMemory(rgb, sizeof(rgb));
}


/*----------------------------------------------------------------------------*/
/* main                                                                       */

int _cdecl main(int argc, char *argv[])
{
    PTCTEMP_CONTEXT  pCtx;
    UNICODE_STRING   us;
    ANSI_STRING      as;
    unsigned         uCipher;
    unsigned         uHash;
    unsigned         uIterationCount;
    int              i;

#if DEBUG
    if (argc && stricmp(argv[argc-1], "dbg") == 0) {
        argc--;
        argv[argc] = NULL;
        DbgBreakPoint();
    }
#endif

    /* Initialization */
    if (!NaInitialize(&pCtx, sizeof(TCTEMP_CONTEXT), MAX_IMAGE_SECTOR_SIZE)) {
        return STATUS_NO_MEMORY;
    }

    TC_InitEntropyPool(&pCtx->ec.ep, 0);
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);
    NaFillEntropyPoolWithSystemInformation(&pCtx->ec,
      pCtx->rgbSystemInformation, sizeof(pCtx->rgbSystemInformation));
    NaFillEntropyPoolWithHardwareRNG(&pCtx->ec,
      pCtx->rgbSystemInformation, 64);
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);

    /* Default settings */
    uCipher = TC_AES;
    uHash = TC_SHA1;
    pCtx->pcwszVolumeName       = L"c:\\tctemp.tc";
    pCtx->ic.chDriveLetter      = L'T';
    pCtx->ic.pcwszImageFileName = L"c:\\tctemp.img";

    /* Command line */
    if (argc > 1) {
        switch(argv[1][0]) {
        case 'A':
            if (argv[1][1] == 'T' && argv[1][2] == 'S') {
                uCipher = TC_AES_TWOFISH_SERPENT;
            } else if (argv[1][1] == 'T') {
                uCipher = TC_AES_TWOFISH;
            } else {
                uCipher = TC_AES;
            }
            break;
        case 'B':
            uCipher = TC_BLOWFISH;
            break;
        case 'C':
            uCipher = TC_CAST5;
            break;
        case 'D':
            uCipher = TC_TRIPLEDES;
            break;
        case 'S':
            if (argv[1][1] == 'T' && argv[1][2] == 'A') {
                uCipher = TC_SERPENT_TWOFISH_AES;
            } else if (argv[1][1] == 'A') {
                uCipher = TC_SERPENT_AES;
            } else {
                uCipher = TC_SERPENT;
            }
            break;
        case 'T':
            if (argv[1][1] == 'S') {
                uCipher = TC_TWOFISH_SERPENT;
            } else {
                uCipher = TC_TWOFISH;
            }
            break;

        }
    }
    if (argc > 2) {
        switch(argv[2][0]) {
        case 'r':
            uHash = TC_RIPEMD160;
            break;
        case 's':
            uHash = TC_SHA1;
            break;
        case 'w':
            uHash = TC_WHIRLPOOL;
            break;
        }
    }
    uIterationCount = (uHash == TC_WHIRLPOOL) ? 1000 : 2000;
    if (argc > 3) {
        int iMul = atoi(argv[3]);
        if (iMul > 1 && iMul <= 2000000) {
            uIterationCount *= (unsigned)iMul;
        } else if (iMul < -(int)uIterationCount) {
            uIterationCount = (unsigned) -iMul;
        }
    }
    if (argc > 4) {
        CHAR ch = (CHAR)toupper(argv[4][0]);
        if (ch >= 'A' && ch <= 'Z') {
            pCtx->ic.chDriveLetter = ch;
        }
    }
    if (argc > 5) {
        RtlInitAnsiString(&as, argv[5]);
        as.MaximumLength = ++as.Length; /* include terminating NUL */
        if (NT_SUCCESS(RtlAnsiStringToUnicodeString(&us, &as, TRUE))) {
            pCtx->pcwszVolumeName = us.Buffer;
        }
    }
    if (argc > 6) {
        RtlInitAnsiString(&as, argv[6]);
        as.MaximumLength = ++as.Length; /* include terminating NUL */
        if (NT_SUCCESS(RtlAnsiStringToUnicodeString(&us, &as, TRUE))) {
            pCtx->ic.pcwszImageFileName = us.Buffer;
        }
    }

#if DEBUG
    for (i = 0; i < argc; i++) {
        PSTR p = (PSTR)argv[i];
        NaPrintf(L"%S ", p);
    }
    NaPrintf(L"\n\n");
#endif

    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);

    NaReadVolumeHeader(pCtx);

    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);
    NaFillEntropyPoolWithFileInfo(&pCtx->ec, pCtx->ic.pcwszImageFileName,
      pCtx->rgbSystemInformation, sizeof(pCtx->rgbSystemInformation));
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);
    NaFillEntropyPoolWithSystemInformation(&pCtx->ec,
      pCtx->rgbSystemInformation, sizeof(pCtx->rgbSystemInformation));
    NaFillEntropyPoolWithTimeStampCounter(&pCtx->ec);

    if (TC_MakeRandomHeader(&pCtx->tch, &pCtx->tcpw,
        &pCtx->ec.ep, uHash, uCipher, uIterationCount, &pCtx->tcb)) {
        if (NaWriteVolumeHeader(pCtx)) {
            if (NaTcOpen(&pCtx->tc)) {
                if (NaTcMountVolume(&pCtx->tc, pCtx->ic.chDriveLetter,
                    pCtx->pcwszVolumeName, pCtx->tcpw.rgbPassword,
                    pCtx->tcpw.cbPasswordSize)) {
#if DEBUG
                    NaPrintf(L"Restoring file system...\n");
#endif
                    if (!NaRestoreFileSystem(&pCtx->ic)) {
                        NaTcDismountVolume(&pCtx->tc,
                            pCtx->ic.chDriveLetter);
                    }
                }
                NaTcClose(&pCtx->tc);
            }
        }
    }

    RtlSecureZeroMemory(pCtx, sizeof(*pCtx));
    NaWipeStack();

    NaUninitialize();

    return 0;
}
